use color_eyre::eyre;
use color_eyre::eyre::WrapErr;
use color_eyre::Help;
use color_eyre::Result;
use color_eyre::SectionExt;
use reqwest::StatusCode;
use serde::Deserialize;
use serde::Serialize;
use serde_json::json;
use std::io;
use std::io::Read;
use std::path::Path;
use url::Url;

use crate::util::make_client;
use crate::Action;

#[derive(Deserialize, Serialize, Debug)]
struct Source {
    access_token: String,
    owner: String,
    url: Url,
    repo: String,
    context: String,
    description: Option<String>,
    target_url: Option<String>,
}

#[derive(Deserialize, Serialize, Debug, Copy, Clone)]
#[serde(rename_all = "snake_case")]
enum State {
    Pending,
    Success,
    Error,
    Failure,
    Warning,
}

impl std::fmt::Display for State {
    fn fmt(&self, f: &mut std::fmt::Formatter<'_>) -> std::fmt::Result {
        match self {
            State::Pending => write!(f, "pending"),
            State::Success => write!(f, "success"),
            State::Error => write!(f, "error"),
            State::Failure => write!(f, "failure"),
            State::Warning => write!(f, "warning"),
        }
    }
}

#[derive(Deserialize, Serialize, Debug)]
struct Params {
    sha: String,
    state: State,
    description: Option<String>,
}

#[derive(Deserialize, Serialize, Debug)]
struct Config {
    /// Resource configuration.
    /// Passed verbatim from the definition in the pipeline.
    source: Source,
    params: Params,
}

fn action_out(conf: Config, _dir: impl AsRef<Path>) -> Result<()> {
    let params = conf.params;
    let description = params.description.or(conf.source.description);

    let client = make_client(&conf.source.access_token)?;
    let url = conf
        .source
        .url
        .join(&format!(
            // The trailing slash is required, to make sure this entire URL is used
            // when `join`ing the file name later. Otherwise, the last component
            // would be considered a "file" part itself, and replaced by a future `join`.
            "api/v1/repos/{}/{}/statuses/{}",
            conf.source.owner, conf.source.repo, params.sha
        ))
        .wrap_err("Invalid URL")?;

    let body = json!({
        "context": conf.source.context,
        "description": description,
        "state": params.state,
        "target_url": conf.source.target_url,
    });

    let res = client
        .post(url.clone())
        .json(&body)
        .send()
        .wrap_err_with(|| format!("Failed to send request 'POST {}'", url))?;

    match res.status() {
        StatusCode::CREATED => {
            eprintln!(
                "Created status '{}' on commit '{}'",
                params.state, params.sha
            );
        }
        StatusCode::BAD_REQUEST => {
            eyre::bail!(
                "Invalid request: {:?}. state={}, context={}, description={:?}, target_url={:?}",
                res.text(),
                params.state,
                conf.source.context,
                description,
                conf.source.target_url
            )
        }
        code => {
            eyre::bail!("Unexpected status code {}\ntext = {:?}", code, res.text());
        }
    }

    println!("{{ \"version\": {{}} }}");
    Ok(())
}

pub(crate) fn run(action: &Action) -> Result<()> {
    // TODO: Gitea does actually support fetching statuses, making `check` and `in` viable,
    // theoretically. But it also doesn't make much sense to implement them.
    match action {
        Action::Check => {
            // Dummy implemented that always reports nothing.
            println!("[]");
            Ok(())
        }
        Action::In { dest: _ } => {
            println!("{{}}");
            Ok(())
        }
        Action::Out { src } => {
            let config: Config = {
                let mut buf = String::new();
                io::stdin()
                    .read_to_string(&mut buf)
                    .wrap_err("Failed to read from stdin")?;

                if buf.is_empty() {
                    eyre::bail!("No data received on stdin");
                }

                serde_json::from_str(&buf)
                    .wrap_err("Failed to parse config")
                    .with_suggestion(|| {
                        "Double-check the `source` and `params` sections in your pipeline"
                    })
                    .with_section(|| buf.header("JSON"))?
            };

            action_out(config, src)
        }
    }
}
